# -*- coding: utf-8 -*-
import json
import uuid

from .BIZ.TL.ST import Story, StoryMedia


def loggedIn(func):
    def checkLogin(*args, **kwargs):
        if args[0].can_use_timeline:
            return func(*args, **kwargs)
        else:
            raise Exception("can't use Timeline func")

    return checkLogin


class Timeline:
    def __init__(self):
        TIMELINE_CHANNEL_ID = "1341209950"
        self.can_use_timeline = False
        try:
            if self.APP_TYPE in ["CHROMEOS"]:
                TIMELINE_CHANNEL_ID = "1341209850"
            self.server.timelineHeaders = {
                "x-line-application": self.server.Headers["x-line-application"],
                "User-Agent": self.server.Headers["User-Agent"],
                "X-Line-Mid": self.mid,
                "X-Line-Access": self.authToken,
                "X-Line-ChannelToken": self.checkAndGetValue(
                    self.approveChannelAndIssueChannelToken(TIMELINE_CHANNEL_ID),
                    "channelAccessToken",
                    5,
                ),
                "x-lal": self.LINE_LANGUAGE,
                "X-LAP": "5",
                "X-LPV": "1",
                "X-LSR": self.LINE_SERVICE_REGION,
                "x-line-bdbtemplateversion": "v1",
                "x-line-global-config": "discover.enable=true; follow.enable=true"
                # "X-Line-PostShare": "true"
                # "X-Line-StoryShare": "true"
            }
            self.can_use_timeline = True
            if self.APP_TYPE in ["CHROMEOS"]:
                self.server.timelineHeaders[
                    "origin"
                ] = "chrome-extension://CHRLINE-v2.5.0-rc-will-not-be-released"

            # BIZ
            self.story = Story(8)

            # OTO
            self.otoMids = {}
            self.syncOtoAccount()
        except Exception as e:
            self.log(f"can't use Timeline: {e}")

    """ TIMELINE """

    @loggedIn
    def getSocialProfileDetail(
        self,
        mid,
        withSocialHomeInfo=True,
        postLimit=10,
        likeLimit=6,
        commentLimit=10,
        storyVersion="v7",
        timelineVersion="v57",
        postId=None,
        updatedTime=None,
    ):
        params = {
            "homeId": mid,
            "withSocialHomeInfo": withSocialHomeInfo,
            "postLimit": postLimit,
            "likeLimit": likeLimit,
            "commentLimit": commentLimit,
            "storyVersion": storyVersion,
            "timelineVersion": timelineVersion,
        }
        if postId is not None:
            # post offset
            params["postId"] = postId
            params["updatedTime"] = updatedTime
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders, {"x-lhm": "GET"}
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/hm/api/v1/home/socialprofile/post.json", params
        )
        r = self.server.postContent(url, headers=hr, data="")
        return r.json()

    @loggedIn
    def getSocialProfileMediaDetail(
        self,
        mid: str,
        withSocialHomeInfo: bool = True,
        postLimit: int = 10,
        withStory: bool = True,
        storyVersion: str = "v7",
        timelineVersion: str = "v57",
        postId: str = None,
        updatedTime: int = None,
    ):
        params = {
            "homeId": mid,
            "withSocialHomeInfo": withSocialHomeInfo,
            "withStory": withStory,
            "postLimit": postLimit,
            "storyVersion": storyVersion,
            "timelineVersion": timelineVersion,
        }
        if postId is not None:
            # post offset
            params["postId"] = postId
            params["updatedTime"] = updatedTime
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders, {"x-lhm": "GET"}
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN,
            "/hm/api/v1/home/socialprofile/mediapost.json",
            params,
        )
        r = self.server.postContent(url, headers=hr, data="")
        return r.json()

    @loggedIn
    def getProfileDetail(
        self, mid, styleMediaVersion="v2", storyVersion="v7", timelineVersion="v57"
    ):
        params = {
            "homeId": mid,
            "styleMediaVersion": styleMediaVersion,
            "storyVersion": storyVersion,
            "timelineVersion": timelineVersion,
            "profileBannerRevision": 0,
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                # why get follow count with this?
                "x-line-global-config": "discover.enable=true; follow.enable=true",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/hm/api/v1/home/profile.json", params
        )
        r = self.server.postContent(url, headers=hr, data="")
        return r.json()

    @loggedIn
    def updateProfileDetail(self, mid):
        params = {}
        data = {
            "homeId": self.mid,
            "styleMediaVersion": "v2",
            "userStyleMedia": {
                "profile": {"displayName": "鴻"},
            },
            "storyShare": "false",
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders, {"x-lhm": "POST"}
        )
        url = self.LINE_HOST_DOMAIN + "/hm/api/v1/home/profile"
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getProfileCoverDetail(self, mid):
        params = {"homeId": mid}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/hm/api/v1/home/cover.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def updateProfileCoverById(
        self, objid: str = None, vObjid: str = None, storyShare: bool = False
    ):
        data = {"homeId": self.mid, "storyShare": storyShare, "meta": {}}  # heh
        if objid is None and vObjid is None:
            raise ValueError("objid is None")
        if objid:
            data["coverObjectId"] = objid
        if vObjid:
            data["videoCoverObjectId"] = vObjid
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        r = self.server.postContent(
            self.LINE_HOST_DOMAIN + "/hm/api/v1/home/cover.json",
            headers=hr,
            data=json.dumps(data),
        )
        return r.json()

    @loggedIn
    def updateProfileCoverById2(self, objId: str):
        params = {
            "coverImageId": objId,
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v41/home/updateCover.json", params
        )
        r = self.server.getContent(url, headers=hr)
        return r.json()

    @loggedIn
    def getOACarousel(self):
        params = {"homeId": "ud9b11771afc46b1d9b6faa646ab3c475"}
        data = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v42/feed/carousel/oa.json", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getPartlyBlockContacts(self):
        params = {"id": self.mid, "limit": 30}
        data = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/tl/mapi/v57/contacts/block/partly.json", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getClosedContacts(self):
        params = {"id": self.mid}
        data = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/tl/mapi/v41/contacts/block", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getHideContacts(self, MID):
        params = {"id": self.mid}
        data = {"mid": "u67c43239c865dfce6addb41c6b3c0edd"}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/tl/mapi/v41/contacts/hide", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getAutoOpenOption(self):
        params = {}
        data = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/tl/mapi/v41/contact/autoopen", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getHideGrouphomeList(self):
        params = {}
        data = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/ma/api/v24/grouphome/hide/list.json", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getNewpostStatus(self):
        params = {}
        data = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/mapi/v41/status/newpost", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getGroupProfileimageList(self):
        params = {}
        data = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v24/group/profileimage/list.json", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getUserProfile(self, mid: str):
        params = {"userMid": mid}
        data = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/ma/api/v1/profile/get.json", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.text

    @loggedIn
    def getUserDetail(self, mid: str):
        params = {"userMid": mid}
        data = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/ma/api/v1/userpopup/getDetail.json", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.text

    @loggedIn
    def syncBuddygroup(self, mid: str):
        params = {"userMid": mid, "lastUpdated": 0}
        data = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/tl/mapi/v41/home/buddygroup/sync", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.text

    @loggedIn
    def sendContactV2(self, homeId, targetMids):
        url = self.LINE_HOST_DOMAIN + "/hm/api/v1/home/profile/share"
        data = {
            "homeId": homeId,
            "shareType": "FLEX_OA_HOME_PROFILE_SHARING",
            "targetMids": targetMids,
        }
        r = self.server.postContent(
            url, headers=self.server.timelineHeaders, data=json.dumps(data)
        )
        result = r.json()
        if result["message"] == "success":
            return result["result"]
        else:
            return False

    @loggedIn
    def getTimelintTab(self, postLimit=20, likeLimit=6, commentLimit=10, requestTime=0):
        url = self.LINE_HOST_DOMAIN + "/tl/api/v57/timeline/tab.json"
        data = {
            "feedRequests": {
                "FEED_LIST": {
                    "version": "v57",
                    "queryParams": {
                        "postLimit": postLimit,
                        "likeLimit": likeLimit,
                        "commentLimit": commentLimit,
                        "requestTime": 0,
                        "userAction": "TAP-REFRESH_UEN",
                        "order": "RANKING",
                    },
                    "requestBody": {
                        "discover": {"contents": ["CP", "PI", "PV", "PL", "LL"]}
                    },
                },
                "STORY": {"version": "v6"},
            }
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "Content-type": "application/json", "x-lsr": "TH"},
        )
        r = self.server.postContent(url, headers=hr, data=json.dumps(data))
        result = r.json()
        if result["message"] == "success":
            return result["result"]
        else:
            return False

    @loggedIn
    def updateCmtLike(self, homeId: str, noti: bool = True):
        """update note notify"""
        params = {
            "homeId": homeId,
        }
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN,
            "/mh/api/v21/grouphome/notisetting/updateCmtLike.json",
            params,
        )
        data = {"notiSet": [{"notiType": "NOTE_CMTLIKE", "noti": noti}]}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getTalkroomStatus(self, userMid: str):
        """check talkroom has updated"""
        params = {
            "userMid": userMid,
        }
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/ma/api/v1/talkroom/get.json", params
        )
        data = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getHomeProfileBridge(self, homeId: str):
        """update note notify"""
        params = {
            "homeId": homeId,
        }
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v1/home/profileBridge.json", params
        )
        data = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    """ POST """

    @loggedIn
    def createPost(
        self,
        homeId: str,
        text: str = None,
        sharedPostId: str = None,
        textSizeMode: str = "NORMAL",
        backgroundColor: str = "#FFFFFF",
        textAnimation: str = "NONE",
        readPermissionType: str = "ALL",
        readPermissionGids: list = None,
        holdingTime: int = None,
        stickerIds: list = None,
        stickerPackageIds: list = None,
        locationLatitudes: list = None,
        locationLongitudes: list = None,
        locationNames: list = None,
        mediaObjectIds: list = None,
        mediaObjectTypes: list = None,
        sourceType: str = "TIMELINE",
    ):
        """
        - readPermissionType:
            ALL,
            FRIEND,
            GROUP,
            EVENT,
            NONE;
        - textSizeMode:
            AUTO,
            NORMAL;
        - textAnimation:
            NONE,
            SLIDE,
            ZOOM,
            BUZZ,
            BOUNCE,
            BLINK;
        """
        if readPermissionGids is None:
            readPermissionGids = []
        if stickerIds is None:
            stickerIds = []
        if stickerPackageIds is None:
            stickerPackageIds = []
        if locationLatitudes is None:
            locationLatitudes = []
        if locationLongitudes is None:
            locationLongitudes = []
        if locationNames is None:
            locationNames = []
        if mediaObjectIds is None:
            mediaObjectIds = []
        if mediaObjectTypes is None:
            mediaObjectTypes = []
        if self.getToType(homeId) == 0:
            self.log(f"Can't use mid {homeId} to create post, will be get oto mid...")
            if homeId in self.otoMids:
                homeId = self.otoMids[homeId]
                self.log(f"Use oto mid {homeId} to create post...")
            else:
                self.log(f"Can't find oto mid: {homeId}, ready to create oto mid...")
                otoRes = self.createOtoAccount(self.mid, homeId)
                otoMid = otoRes.get("result", {}).get("groupId")
                if otoMid is None:
                    raise ValueError(f"Oto Mid is None: {otoRes}")
                self.log(f"Created oto mid {otoMid} to create post...")
                self.otoMids[homeId] = otoMid
                homeId = otoMid
        params = {"homeId": homeId, "sourceType": sourceType}
        postInfo = {
            "readPermission": {"type": readPermissionType, "gids": readPermissionGids},
        }
        stickers = []
        locations = []
        medias = []
        for stickerIndex, stickerId in enumerate(stickerIds):
            stickers.append(
                {
                    "id": stickerId,
                    "packageId": stickerPackageIds[stickerIndex],
                    "packageVersion": 1,
                    "hasAnimation": True,  # TODO: check it
                    "hasSound": True,  # TODO: check it
                    "stickerResourceType": "ANIMATION",  # TODO: check it
                }
            )
        for locatioIndex, locationLatitude in enumerate(locationLatitudes):
            locations.append(
                {
                    "latitude": locationLatitude,
                    "longitude": locationLongitudes[locatioIndex],
                    "name": locationNames[locatioIndex],
                }
            )
        for mediaIndex, mediaObjectId in enumerate(mediaObjectIds):
            medias.append(
                {
                    "objectId": mediaObjectId,
                    "type": mediaObjectTypes[mediaIndex],
                    # "width": 1016,
                    # "height": 453,
                    # "size": 16022,
                    "obsFace": "[]",
                }
            )
        contents = {
            "contentsStyle": {
                "textStyle": {
                    "textSizeMode": textSizeMode,
                    "backgroundColor": backgroundColor,
                    "textAnimation": textAnimation,
                },
                "mediaStyle": {"displayType": "GRID_1_A"},
            },
            "stickers": stickers,
            "locations": locations,
            "media": medias,
        }
        if holdingTime is not None:
            postInfo["holdingTime"] = holdingTime
        if text is not None:
            contents["text"] = text
        if sharedPostId is not None:
            contents["sharedPostId"] = sharedPostId
        data = {"postInfo": postInfo, "contents": contents}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v57/post/create.json", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def updatePost(self, homeId: str, postData: dict, sourceType: str = "TIMELINE"):
        params = {"homeId": homeId, "sourceType": sourceType}
        data = postData.copy()
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v57/post/update.json", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def deletePost(self, homeId: str, postId: str):
        params = {"homeId": homeId, "postId": postId}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v57/post/delete.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def getPost(self, mid: str, postId: str, sourceType: str = "MYHOME"):
        params = {"homeId": mid, "postId": postId, "sourceType": sourceType}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v57/post/get.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def listPost(
        self,
        mid: str,
        postId: str = None,
        updatedTime: int = None,
        sourceType: str = "TALKROOM",
    ):
        """list posts for chat room.
        use postId and updatedTime to fetch next posts"""
        params = {"homeId": mid, "sourceType": sourceType}
        if postId is not None:
            params["postId"] = postId
        if updatedTime is not None:
            params["updatedTime"] = updatedTime
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v57/post/list.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def createComment(
        self,
        mid: str,
        contentId: str,
        text: str = "",
        stickerId: int = None,
        stickerPackageId: int = None,
        mediaObjectId: str = None,
        mediaObjectType: str = "PHOTO",
        mediaObjectFrom: str = "cmt",
        sourceType: str = "TIMELINE",
    ):
        params = {"homeId": mid, "sourceType": sourceType}
        contentsList = []
        if stickerId is not None:
            contentsList.append(
                {
                    "categoryId": "sticker",
                    "extData": {
                        "id": stickerId,
                        "packageId": stickerPackageId,
                        "packageVersion": 1,
                        "hasAnimation": True,  # TODO: check it
                        "hasSound": True,  # TODO: check it
                        "stickerResourceType": "ANIMATION",  # TODO: check it
                    },
                }
            )
        if mediaObjectId is not None:
            contentsList.append(
                {
                    "categoryId": "media",
                    "extData": {
                        "objectId": mediaObjectId,
                        "type": mediaObjectType,
                        "width": 999999999999999,
                        "height": 999999999999999,
                        "size": -1,
                        "obsNamespace": mediaObjectFrom,
                        "serviceName": "myhome",
                    },
                }
            )
        data = {
            "contentId": contentId,
            "commentText": text,
            # "secret": False,  # WTF THIS?
            "contentsList": contentsList,
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v57/comment/create.json", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def deleteComment(self, mid: str, commentId: str):
        params = {"homeId": mid, "commentId": commentId}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v57/comment/delete.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def listComment(self, mid: str, contentId: str, scrollId: str = None):
        params = {
            "homeId": mid,
            # 'actorId': actorId,
            "contentId": contentId,
            # 'limit': 10
        }
        if scrollId is not None:
            params["scrollId"] = scrollId
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
                "x-lpv": "1",
                "x-lsr": "TH",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v57/comment/getList.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def createLike(
        self,
        mid: str,
        contentId: str,
        likeType: str = "1003",
        sharable: bool = False,
        sourceType: str = "TIMELINE",
    ):
        params = {"sourceType": sourceType}
        data = {
            "contentId": contentId,
            "actorId": mid,
            "likeType": likeType,
            "sharable": sharable,
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v57/like/create.json", params
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def cancelLike(self, contentId: str, sourceType: str = "TIMELINE"):
        params = {"contentId": contentId, "sourceType": sourceType}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v41/like/cancel.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def listLike(
        self,
        mid: str,
        contentId: str,
        scrollId: str = None,
        includes: list = None,
        filterType: str = None,
    ):
        if includes is None:
            includes = ["ALL", "GROUPED", "STATS"]
        params = {"homeId": mid, "contentId": contentId, "includes": ",".join(includes)}
        if scrollId is not None:
            params["scrollId"] = scrollId
        if filterType is not None:
            params["filterType"] = filterType  # eg. 1003 for GROUPED
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "Content-type": "application/json",
                "x-lpv": "1",
                "x-lsr": "TH",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v41/like/getList.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def searchNote(self, mid, text):
        data = {
            "query": text,
            "queryType": "TEXT",
            "homeId": mid,
            "postLimit": 20,
            "commandId": 16,
            "channelId": "1341209850",
            "commandType": 188259,
        }
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v46/search/note.json", {}
        )
        r = self.server.postContent(
            url, headers=self.server.timelineHeaders, data=json.dumps(data)
        )
        res = r.json()
        return res["result"]["feeds"]

    @loggedIn
    def sendPostToTalk(self, postId: str, receiveMids: list):
        data = {"postId": postId, "receiveMids": receiveMids}
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v57/post/sendPostToTalk.json", {}
        )
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        r = self.server.postContent(url, headers=hr, data=json.dumps(data))
        return r.json()

    @loggedIn
    def getHashtagPosts(self, query, homeId=None, scrollId=None, range=None):
        if range is None:
            range = ["GROUP"]
        # range: GROUP or unset
        data = {"query": query, "homeId": homeId, "scrollId": scrollId, "range": range}
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v52/hashtag/posts.json", {}
        )
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
                "x-lal": self.LINE_LANGUAGE,
            },
        )
        r = self.server.postContent(url, headers=hr, data=json.dumps(data))
        return r.json()

    @loggedIn
    def getHashtagSuggest(self, query):
        data = {"query": query}
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v52/hashtag/suggest.json", {}
        )
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
                "x-lal": self.LINE_LANGUAGE,
            },
        )
        r = self.server.postContent(url, headers=hr, data=json.dumps(data))
        return r.json()

    @loggedIn
    def getHashtagPopular(self, homeId, limit=20):
        data = {"homeId": homeId, "limit": limit}
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v52/hashtag/popular.json", {}
        )
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
                "x-lal": self.LINE_LANGUAGE,
            },
        )
        r = self.server.postContent(url, headers=hr, data=json.dumps(data))
        return r.json()

    @loggedIn
    def getTimelineUrl(self, homeId):
        data = {"homeId": homeId}
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v55/web/getUrl.json", data
        )
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders, {"x-lal": self.LINE_LANGUAGE}
        )
        r = self.server.getContent(url, headers=hr)
        return r.json()

    @loggedIn
    def getPostShareLink(self, postId):
        data = {"postId": postId}
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/api/v57/post/getShareLink.json", data
        )
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders, {"x-lal": self.LINE_LANGUAGE}
        )
        r = self.server.getContent(url, headers=hr)
        return r.json()

    @loggedIn
    def getDiscoverRecommendFeeds(self, sourcePostId, contents=None):
        if contents is None:
            contents = ["CP", "PI", "PV", "PL", "AD"]
        data = {"sourcePostId": sourcePostId, "contents": contents}
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/tl/discover/api/v1/recommendFeeds", {}
        )
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders, {"x-lal": self.LINE_LANGUAGE}
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    """ ALBUM """

    @loggedIn
    def changeGroupAlbumName(self, mid, albumId, name):
        data = json.dumps({"title": name})
        params = {"homeId": mid}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "PUT",
                "Content-type": "application/json",
                "x-lpv": "1",  # needless
                "x-lsr": "TH",  # needless
                "x-u": "",  # needless
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/ext/album/api/v3/album/%s" % albumId, params
        )
        r = self.server.postContent(url, data=data, headers=hr)
        # r.json()['code'] == 0: success
        return r.json()

    @loggedIn
    def deleteGroupAlbum(self, mid, albumId):
        params = {"homeId": mid}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "DELETE",
                "Content-type": "application/json",
                "x-lpv": "1",  # needless
                "x-lsr": "TH",  # needless
                "x-u": "",  # needless
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/ext/album/api/v4/album/%s" % albumId, params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def addImageToAlbum(self, mid: str, albumId: str, oid: any):
        """
        Add an image to the album

        oid is a string, you can use list for multiple oids
        """
        oids = []
        if type(oid) is str:
            oid = [oid]
        if type(oid) is list:
            for _oid in oid:
                oids.append({"oid": _oid})
        else:
            raise ValueError(f"Not support oid: {oid}")
        params = {"homeId": mid}
        data = json.dumps({"photos": oids})

        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "PUT",
                "content-type": "application/json",
                # change it if you want update many images
                "x-album-stats": "e2FsYnVtUGhvdG9BZGRDb3VudD0xfQ==",
            },
        )

        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/ext/album/api/v3/photos/%s" % albumId, params
        )
        r = self.server.postContent(url, data=data, headers=hr)
        return r.json()

    @loggedIn
    def getAlbumImages(self, mid, albumId):
        params = {"homeId": mid}

        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "content-type": "application/json",
            },
        )

        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/ext/album/api/v3/photos/%s" % albumId, params
        )
        r = self.server.postContent(url, headers=hr)

        return r.json()

    @loggedIn
    def deleteAlbumImages(self, mid: str, albumId: str, id: any):
        """
        Delete an image on the album

        id is a string, you can use list for multiple ids
        * id is photo id, not objId
        """
        ids = []
        if type(id) is str:
            id = [id]
        if type(id) is list:
            for _id in id:
                ids.append({"id": _id})
        else:
            raise ValueError(f"Not support id: {id}")
        params = {"homeId": mid}
        data = json.dumps({"photos": ids})

        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "content-type": "application/json"},
        )

        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "ext/album/api/v3/photos/delete/%s" % albumId, params
        )
        r = self.server.postContent(url, data=data, headers=hr)
        # {"code":0,"message":"success","result":true}
        return r.json()

    @loggedIn
    def getAlbums(self, homeId):
        params = {"homeId": homeId}
        data = {}

        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )

        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/ext/album/api/v3/albums", params
        )
        r = self.server.postContent(url, json=data, headers=hr)
        return r.json()

    @loggedIn
    def getAlbumUsers(self, mid, albumId):
        params = {"homeId": mid}
        data = {}

        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )

        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/ext/album/api/v3/users/%s" % albumId, params
        )
        r = self.server.postContent(url, json=data, headers=hr)
        return r.json()

    @loggedIn
    def createGroupAlbum(self, mid, name):
        data = json.dumps({"title": name, "type": "image"})
        params = {"homeId": mid, "count": "1", "auto": "1"}
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/ext/album/api/v3/album", params
        )
        r = self.server.postContent(url, data=data, headers=self.server.timelineHeaders)
        if r.status_code != 201:
            print(r.text)
        return r.json()

    """ STORY """

    @loggedIn
    def createStoryContent(
        self, oid: str, ohash: str, oType: str = "IMAGE", shareType: str = "PUBLIC"
    ):
        """
        Create story content.

        - shareType:
            PUBLIC,
            FRIEND,
            GROUP,
            FOLLOW_GROUP;

        """
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "content-type": "application/json"},
        )
        medias = [StoryMedia(oid, ohash, oType)]
        data = self.story.CreateStoryContent(
            **{
                "sourceType": "USER",
                "contentType": "USER",
                "medias": medias,
                "shareType": shareType,
            }
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, self.story.GetApiPath("story/content/create")
        )
        r = self.server.postContent(url, json=data, headers=hr)
        return r.json()

    @loggedIn
    def deleteStoryContent(self, contentId: str):
        """Delete story content."""
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "content-type": "application/json"},
        )
        data = self.story.DeleteStoryContent(**{"contentId": contentId})
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, self.story.GetApiPath("story/content/delete")
        )
        r = self.server.postContent(url, json=data, headers=hr)
        return r.json()

    @loggedIn
    def getStoryList(self, userMids: list):
        """Get story list by mids."""
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "content-type": "application/json"},
        )
        data = self.story.GetStoryList(**{"userMid": self.mid, "userMids": userMids})
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, self.story.GetApiPath("story/list")
        )
        r = self.server.postContent(url, json=data, headers=hr)
        return r.json()

    @loggedIn
    def getStory(self, userMid: str):
        """Get story by mid."""
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "content-type": "application/json"},
        )
        data = self.story.GetStory(**{"userMid": userMid})
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, self.story.GetApiPath("story")
        )
        r = self.server.postContent(url, json=data, headers=hr)
        return r.json()

    @loggedIn
    def getStoryContent(self, contentId: str):
        """Get story content."""
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "content-type": "application/json"},
        )
        data = self.story.GetStoryContent(**{"contentId": contentId})
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, self.story.GetApiPath("story/content")
        )
        r = self.server.postContent(url, json=data, headers=hr)
        return r.json()

    @loggedIn
    def readStoryContent(self, contentId: str):
        """Read story content."""
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "content-type": "application/json"},
        )
        data = self.story.ReadStoryContent(**{"contentId": contentId, "tsId": ""})
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, self.story.GetApiPath("story/content/read")
        )
        r = self.server.postContent(url, json=data, headers=hr)
        return r.json()

    @loggedIn
    def likeStoryContent(
        self, contentId: str, tsId: str, like: bool, likeType: str = "1001"
    ):
        """Like story content."""
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "content-type": "application/json"},
        )
        data = self.story.LikeStoryContent(
            **{"contentId": contentId, "tsId": tsId, "like": like, "likeType": likeType}
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, self.story.GetApiPath("story/content/like")
        )
        r = self.server.postContent(url, json=data, headers=hr)
        return r.json()

    @loggedIn
    def getStoryContentLikeList(
        self,
        contentId: str,
        include: str = "MERGED",
        likeType: str = "1001",
        size: int = 20,
        scrollId: str = None,
    ):
        """Get story content like list."""
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "content-type": "application/json"},
        )
        req = {
            "contentId": contentId,
            "include": include,
            "likeType": likeType,  # opt
            "size": size,
        }
        if scrollId is not None:
            req["scrollId"] = scrollId
        data = self.story.StoryContentLikeList(**req)
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, self.story.GetApiPath("story/content/like/list")
        )
        r = self.server.postContent(url, json=data, headers=hr)
        return r.json()

    @loggedIn
    def sendMessageByStory(
        self, toMid: str, contentId: str, message: str, tsId: str = ""
    ):
        """Send message by story."""
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "content-type": "application/json"},
        )
        req = {
            "to": {"userMid": toMid, "tsId": tsId},
            "contentId": contentId,
            "message": message,
        }
        data = self.story.SendStoryMessage(**req)
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, self.story.GetApiPath("story/message/send")
        )
        r = self.server.postContent(url, data=data, headers=hr)
        return r.json()

    @loggedIn
    def getRecentStoryList(self, lastRequestTime: int, lastTimelineVisitTime: int = 0):
        """Get recent story list."""
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "content-type": "application/json"},
        )
        data = self.story.GetRecentStoryList(
            **{
                "lastRequestTime": lastRequestTime,
                "lastTimelineVisitTime": lastTimelineVisitTime,
            }
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, self.story.GetApiPath("story/recentstory/list")
        )
        r = self.server.postContent(url, json=data, headers=hr)
        return r.json()

    @loggedIn
    def getNewStory(self, lastTimelineVisitTime: int = 0):
        """Get new story."""
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "content-type": "application/json"},
        )
        data = self.story.GetNewStory(
            **{"lastTimelineVisitTime": lastTimelineVisitTime}
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, self.story.GetApiPath("story/newstory")
        )
        r = self.server.postContent(url, json=data, headers=hr)
        return r.json()

    @loggedIn
    def fetchNewNotification(self, revision: int = 0):
        params = {"serviceCode": "TL", "revision": revision}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "X-Line-TL-Upstream-ID": "1583881852"},
        )
        data = {}
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/eg/lin/api/v1/notification/fetch/new", params
        )
        r = self.server.postContent(url, json=data, headers=hr)
        return r.json()

    """ Search """

    @loggedIn
    def lnexearch(self, q):
        params = {"q": q}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                # x-line-channeltoken: 1557852768
            },
        )
        url = self.server.urlEncode(self.LINE_SEARCH_DOMAIN, "/lnexearch", params)
        r = self.server.postContent(url, headers=hr)
        return r.json()

    """ Sc """

    def getPageInfo(self, url):
        params = {"url": url, "caller": "TALK", "lang": self.LINE_LANGUAGE}
        hr = self.server.additionalHeaders(
            # self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/sc/api/v2/pageinfo/get.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    """ One to one """

    @loggedIn
    def syncOtoAccount(self, userMid: str = None):
        """Sync oto accounts."""
        if userMid is None:
            userMid = self.mid
        params = {"userMid": userMid}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v57/otoaccount/sync.json", params
        )
        r = self.server.postContent(url, headers=hr)
        resp = r.json()
        if userMid == self.mid:
            # update oto accounts cache.
            result = resp.get("result")
            if result is None:
                self.log(f"Can't update oto accounts.")
            else:
                updated = result.get("updated", [])
                for i in updated:
                    groupId = i.get("groupId")
                    userMid = i.get("userMid")
                    self.otoMids[userMid] = groupId
                self.log(f"Oto accounts updated.", True)
        return resp

    @loggedIn
    def createOtoAccount(self, userMid: str, friendMid: str):
        """Create oto account."""
        params = {"userMid": userMid, "friendMid": friendMid}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "POST", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/mh/api/v57/otoaccount/create.json", params
        )
        r = self.server.postContent(url, json=params, headers=hr)
        return r.json()

    """ Keep """

    @loggedIn
    def syncKeep(self, revision=0, limit=30):
        params = {"revision": revision, "limit": limit}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/kp/api/v27/keep/sync.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def fetchKeep(self, startRevision: int = 0, endRevision: int = 0, limit: int = 30):
        params = {
            "startRevision": startRevision,
            "endRevision": endRevision,
            "limit": limit,
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "GET",
                "content-type": "application/json",
                "x-source": "KEEP-PICKER",  # or KEEP-DEFAULT
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/kp/api/v27/keep/fetch.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def createKeepContent(self):
        raise NotImplementedError("createKeepContent is not implemented")
        params = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/kp/api/v27/keep/create.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def updateKeepContent(self):
        raise NotImplementedError("updateKeepContent is not implemented")
        params = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/kp/api/v27/keep/update.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def deleteKeepContent(self):
        raise NotImplementedError("deleteKeepContent is not implemented")
        params = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/kp/api/v27/keep/delete.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def getKeepSize(self, revision: int = 0, limit: int = 30):
        params = {"revision": revision, "limit": limit}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/kp/api/v27/keep/size.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def initKeepStatus(self):
        params = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/kp/api/v27/keep/init.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def deleteKeepObs(self):
        raise NotImplementedError("deleteKeepObs is not implemented")
        params = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/kp/api/v27/keep/obs/delete.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def getKeep(self):
        params = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/kp/api/v27/keep/get.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def deleteKeepMessage(self):
        raise NotImplementedError("deleteKeepMessage is not implemented")
        params = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/kp/api/v27/keep/message/delete.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def pinKeepContents(self):
        raise NotImplementedError("pinKeepContents is not implemented")
        params = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/kp/api/v27/keep/contents/pin.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    @loggedIn
    def unpinKeepContents(self):
        raise NotImplementedError("UnpinKeepContents is not implemented")
        params = {}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {"x-lhm": "GET", "content-type": "application/json"},
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN, "/kp/api/v27/keep/contents/unpin.json", params
        )
        r = self.server.postContent(url, headers=hr)
        return r.json()

    """ GroupCall YotTube """

    @loggedIn
    def getYouTubeVideos(self, videoIds: list):
        params = {}
        data = {
            "id": ",".join(videoIds),
            "part": "snippet,contentDetails,id,liveStreamingDetails,status,statistics",
            "fields": "items(snippet(publishedAt,title,thumbnails(default,high,medium),channelTitle,"
            "liveBroadcastContent),contentDetails(duration,contentRating),id,liveStreamingDetails("
            "concurrentViewers,scheduledStartTime),status(embeddable),statistics(viewCount, commentCount)) ",
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "content-type": "application/json",
                "x-voip-service-id": "gc",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN + self.LINE_VOIP_GROUP_CALL_YOUTUBE_ENDPOINT,
            "/api/videos",
            params,
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getYouTubeVideosWithQuery(self, query="bao", pageToken=None):
        params = {}
        data = {
            "q": query,
            "part": "id",
            "fields": "nextPageToken,items(id/videoId)",
            "safeSearch": "strict",
            "order": "relevance",
            "maxResults": 50,
            "type": "video",
        }
        if pageToken is not None:
            data["pageToken"] = pageToken
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "content-type": "application/json",
                "x-voip-service-id": "gc",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN + self.LINE_VOIP_GROUP_CALL_YOUTUBE_ENDPOINT,
            "/api/search",
            params,
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getYouTubeVideosWithPopular(self, pageToken=None):
        params = {}
        data = {
            "chart": "mostPopular",
            "part": "snippet,contentDetails,id,liveStreamingDetails,status,statistics",
            "fields": "items(snippet(publishedAt,title,thumbnails(default,high,medium),channelTitle,liveBroadcastContent),contentDetails(duration,contentRating),id,liveStreamingDetails(concurrentViewers,scheduledStartTime),status(embeddable),statistics(viewCount))",
            "regionCode": "TH",
            "maxResults": 50,
        }
        if pageToken is not None:
            data["pageToken"] = pageToken
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "content-type": "application/json",
                "x-voip-service-id": "gc",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN + self.LINE_VOIP_GROUP_CALL_YOUTUBE_ENDPOINT,
            "/api/videos",
            params,
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getYouTubeVideosWithPlaylists(self, ids, pageToken=None):
        params = {}
        data = {
            "id": ",".join(ids),
            "part": "id,contentDetails, snippet",
            "fields": "items(contentDetails,id,snippet)",
            "regionCode": "TH",
            "maxResults": 50,
        }
        if pageToken is not None:
            data["pageToken"] = pageToken
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "content-type": "application/json",
                "x-voip-service-id": "gc",
            },
        )
        url = self.server.urlEncode(
            self.LINE_HOST_DOMAIN + self.LINE_VOIP_GROUP_CALL_YOUTUBE_ENDPOINT,
            "/api/playlists",
            params,
        )
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    """ Share List Service """

    @loggedIn
    def createShareList(self, ownerMid: str, members: list):
        data = {
            "ownerMid": ownerMid,
            "name": "????",
            "members": members,
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        url = self.LINE_HOST_DOMAIN + "/ext/timeline/tlgw/sl/api/v2/sharelist/create"
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    """ BDB """

    @loggedIn
    def incrBDBCelebrate(self, boardId: str, incrCnt: int = 1):
        data = {
            "boardId": boardId,
            "from": "POST",
            "incrCnt": incrCnt,
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        url = self.LINE_HOST_DOMAIN + "/tl/api/v1/bdb/celebrate/incr"
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def cancelBDBCelebrate(self, boardId: str):
        data = {
            "boardId": boardId,
            "from": "POST",
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        url = self.LINE_HOST_DOMAIN + "/tl/api/v1/bdb/celebrate/cancel"
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def getBDBBoard(self, boardId: str):
        data = {
            "boardId": boardId,
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        url = self.LINE_HOST_DOMAIN + "/tl/api/v1/bdb/board/get"
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def likeBDBCard(self, boardId: str, cardId: str):
        data = {
            "boardId": boardId,
            "cardId": cardId,
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        url = self.LINE_HOST_DOMAIN + "/tl/api/v1/bdb/card/like/create"
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def unlikeBDBCard(self, boardId: str, cardId: str):
        data = {
            "boardId": boardId,
            "cardId": cardId,
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        url = self.LINE_HOST_DOMAIN + "/tl/api/v1/bdb/card/like/cancel"
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    @loggedIn
    def createBDBCard(
        self, boardId: str, celebratorMid: str, text: str, cardStatus: str = "NORMAL"
    ):
        data = {
            "boardId": boardId,
            "cardStatus": cardStatus,  # NORMAL or HIDDEN
            "celebratorMid": celebratorMid,  # self mid, but why? 🤔
            "text": text,
            "from": "BOARD",
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
            },
        )
        url = self.LINE_HOST_DOMAIN + "/tl/api/v1/bdb/card/create"
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    """ Avatar Service """

    @loggedIn
    def shareAvatarToTalkById(self, avatar_id: str, toMids: list):
        data = {"avatar_id": avatar_id, "to": toMids}
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
                # 'X-Line-Clientid': ''
            },
        )
        url = self.LINE_HOST_DOMAIN + "/ex/ya/am/v1/share"
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()

    """ Translate Service """

    @loggedIn
    def translate(self, originalText: str, sLang: str = "auto", tLang: str = "th"):
        data = {
            "originalText": originalText,
            "sLang": sLang,
            "tLang": tLang,
            "xMode": 0,
            "id": str(uuid.uuid1()),
            "service": "talk",
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-lhm": "POST",
                "Content-type": "application/json",
                "x-client-channel": "chat_context",
                "X-Line-Translate-From": "line_android",
            },
        )
        url = self.LINE_HOST_DOMAIN + "/ds/translate/legyTransAPI.nhn"
        r = self.server.postContent(url, headers=hr, json=data)
        return r.json()



# -*- coding: utf-8 -*-


class BuddyService(object):
    BuddyService_REQ_TYPE = 3
    BuddyService_RES_TYPE = 3
    BuddyService_API_PATH = None

    def __init__(self):
        self.BuddyService_API_PATH = self.LINE_BUDDY_ENDPOINT

    def getPromotedBuddyContacts(self, language="th_TH", country="TH"):
        METHOD_NAME = "getPromotedBuddyContacts"
        sqrd = [128, 1, 0, 1] + \
            self.getStringBytes(METHOD_NAME) + [0, 0, 0, 0]
        sqrd += [11, 0, 2] + self.getStringBytes(language)
        sqrd += [11, 0, 3] + self.getStringBytes(country)
        sqrd += [0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_BUDDY_ENDPOINT, sqrd, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyCategoryView(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getBuddyCategoryView is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getBuddyCategoryView", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyChatBarV2(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getBuddyChatBarV2 is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getBuddyChatBarV2", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyChatBar(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getBuddyChatBar is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getBuddyChatBar", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getCountriesHavingBuddy(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getCountriesHavingBuddy is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getCountriesHavingBuddy", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getPopularBuddyBanner(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getPopularBuddyBanner is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getPopularBuddyBanner", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyStatusBarV2(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getBuddyStatusBarV2 is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getBuddyStatusBarV2", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyDetailWithPersonal(self, buddyMid: str, attributeSet: list):
        METHOD_NAME = "getBuddyDetailWithPersonal"
        params = [
            [11, 1, buddyMid],
            [14, 2, [11, attributeSet]]
        ]
        sqrd = self.generateDummyProtocol(
            METHOD_NAME, params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyLive(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getBuddyLive is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getBuddyLive", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyContacts(self, language: str, country: str, classification: str, fromIndex: int, count: int):
        METHOD_NAME = "getBuddyContacts"
        params = [
            [11, 2, language],
            [11, 3, country],
            [11, 4, classification],
            [10, 5, fromIndex],
            [8, 6, count],
        ]
        sqrd = self.generateDummyProtocol(
            METHOD_NAME, params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyTopView(self, language: str, country: str):
        """REMOVED"""
        METHOD_NAME = "getBuddyTopView"
        params = [
            [11, 2, language],
            [11, 3, country],
        ]
        sqrd = self.generateDummyProtocol(
            METHOD_NAME, params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyCollectionEntries(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getBuddyCollectionEntries is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getBuddyCollectionEntries", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getPopularBuddyLists(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getPopularBuddyLists is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getPopularBuddyLists", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getNewlyReleasedBuddyIds(self):
        METHOD_NAME = "getNewlyReleasedBuddyIds"
        params = []
        sqrd = self.generateDummyProtocol(
            METHOD_NAME, params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyOnAir(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getBuddyOnAir is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getBuddyOnAir", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyNewsView(self, language: str, country: str, fromIndex: int, count: int):
        """REMOVED"""
        METHOD_NAME = "getBuddyNewsView"
        params = [
            [11, 2, language],
            [11, 3, country],
            [10, 4, fromIndex],
            [8, 5, count],
        ]
        sqrd = self.generateDummyProtocol(
            METHOD_NAME, params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getCountriesServingOfficialAccountPromotionV2(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception(
            "getCountriesServingOfficialAccountPromotionV2 is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getCountriesServingOfficialAccountPromotionV2", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyStatusBar(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getBuddyStatusBar is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getBuddyStatusBar", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getRichMenuContents(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getRichMenuContents is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getRichMenuContents", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyDetail(self, buddyMid: str):
        METHOD_NAME = "getBuddyDetail"
        params = [
            [11, 4, buddyMid]
        ]
        sqrd = self.generateDummyProtocol(
            METHOD_NAME, params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def findBuddyContactsByQuery(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("findBuddyContactsByQuery is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "findBuddyContactsByQuery", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getBuddyProfilePopup(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getBuddyProfilePopup is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getBuddyProfilePopup", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")

    def getLatestBuddyNewsTimestamp(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getLatestBuddyNewsTimestamp is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getLatestBuddyNewsTimestamp", params, self.BuddyService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.BuddyService_API_PATH, sqrd, self.BuddyService_RES_TYPE, readWith=f"BuddyService.{METHOD_NAME}")



# -*- coding: utf-8 -*-

class HookUtility(object):

    def __init__(self):
        pass

    def checkCmd(self, **kwargs):
        print(f'[checkCmd] {kwargs}')
        return True

    def getPrefix(self, text: str):
        for x in self.prefixes:
            if text.startswith(x):
                return x
        return None

    def genHelp(self, prefix: str = None, user_mid: str = None, msg: any = None):
        '''[C:] Make help by functions'''
        if prefix is None:
            prefix = self.prefixes[0]
        # Make CmdList
        ls = [[x.__name__.replace("_", " "), x.__doc__, x.prefixes,
               x.permissions, x.toType] for x in self.cmdFuncs]
        ls = [l for l in ls if l[1] != None]
        if user_mid is not None:
            ls = [l for l in ls if self.checkPermissions(user_mid, l[3])]
        if msg is not None:
            toType = self.cl.checkAndGetValue(msg, 'toType', 3)
            ls = [l for l in ls if toType in l[4]]
        # Put Prefix
        ls2 = [[prefix+l[0], l[1]] if l[2] else [l[0], l[1]] for l in ls]
        text = '指令表如下:'
        for _cmd in ls2:
            text += f"\n{_cmd[0]}: {_cmd[1]}"
        return text

    def checkPermissions(self, mid: str, permissions: list, excludePermissions: list = []):
        if len(permissions) == 0:
            return True
        for permission in permissions:
            _p = self.getPermission(permission)
            if mid in _p:
                return True
        return False

    def getPermission(self, permission: str):
        return self.db.getData(str(permission).lower(), [])

    def addPermission(self, mid: str, permission: str):
        _k = str(permission).lower()
        _p = self.db.getData(_k, [])
        if mid not in _p:
            _p.append(mid)
            self.db.saveData(_k, _p)
            return True
        return False

    def removePermission(self, mid: str, permission: str):
        _k = str(permission).lower()
        _p = self.db.getData(_k, [])
        if mid in _p:
            _p.remove(mid)
            self.db.saveData(_k, _p)
            return True
        return False

    def getArgs(self, text: str, splitchar: str = ":", defVal: any = None, maxSplit: int = -1):
        args = text.split(splitchar, maxSplit)
        if len(args) > 1:
            return args[1:]
        if defVal is not None:
            return [defVal]
        return None



# -*- coding: utf-8 -*-
import json
import sqlite3

def ezLocker(func):
    def check(*args, **kwargs):
        if args[0]._checkLock():
            return func(*args, **kwargs)
        else:
            raise Exception("can't use Timeline func")
    return check

class BaseDatabase:

    def __init__(self, cl, db_name):
        self.cl = cl
        if db_name is None:
            db_name = self.cl.mid
        self.db_name = db_name
        print("Initialize database...")
        self._ezLocker = False
        self.loadDatabase()
    
    def _lock(self):
        self._ezLocker = True
    
    def _lockRrelease(self):
        self._ezLocker = False
    
    def _checkLock(self):
        while self._ezLocker:
            time.sleep(1 / 1000)
        return True

class SqliteDatabase(BaseDatabase):

    def loadDatabase(self):
        print(f"Loading db_{self.db_name} with Sqli")
        _data = self.cl.getCacheData(".database", f"{self.db_name}.db", pathOnly=True)
        self._sqlite = sqlite3.connect(_data)
        if self._sqlite:
            self._init()
        else:
            raise Exception('can not load database.')
        
    @ezLocker
    def getData(self, _key, _defVal=None):
        SQLITE3_SELECT_CMD = f""" SELECT * FROM [CHRLINE] WHERE [key]=?; """
        cursor = self._sqlite.cursor()
        cursor.execute(SQLITE3_SELECT_CMD, (_key, ))
        rows = cursor.fetchone()
        return _defVal if not rows else self._val2obj(rows[2])

    def saveData(self, _key, _val):
        SQLITE3_SAVE_DATA_CMD = """
            INSERT OR REPLACE INTO [CHRLINE] ([key], [val]) values (?, ?); 
        """
        cursor = self._sqlite.cursor()
        cursor.execute(SQLITE3_SAVE_DATA_CMD, (_key, self._obj2val(_val)))
        self.saveDatabase()

    def saveDatabase(self):
        self._sqlite.commit()
    
    def _obj2val(self, obj):
        a = type(obj)
        if a in [list, dict]:
            return json.dumps(obj)
        elif a in [str, int]:
            return str(obj)
        raise ValueError(f'can not conv data type to string: {a} -> {obj}')
    
    def _val2obj(self, val):
        # TODO: check and reload with json.loads if it is json data
        try:
            val = json.loads(val)
        except:
            print(f"[_val2obj] cant conv data for json: {val}")
        return val
        
    def _init(self):
        SQLITE3_CHRLINE_UNIT_TABLE_CMD = """
            CREATE TABLE IF NOT EXISTS CHRLINE (
                id integer PRIMARY KEY,
                key text NOT NULL UNIQUE,
                val text
            );
        """
        cursor = self._sqlite.cursor()
        cursor.execute(SQLITE3_CHRLINE_UNIT_TABLE_CMD)

class JsonDatabase(BaseDatabase):

    def loadDatabase(self):
        print(f"Loading db_{self.db_name} with Json")
        _data = self.cl.getCacheData(".database", f"{self.db_name}.json")
        self._json = json.loads(_data) if _data is not None and _data != '' else {}

    def getData(self, _key, _defVal=None):
        return self._json.get(_key, _defVal)

    def saveData(self, _key, _val):
        self._json[_key] = _val
        self.saveDatabase()

    def saveDatabase(self):
        self.cl.saveCacheData(".database", f"{self.db_name}.json", json.dumps(self._json))
        print(f"Save db_{self.db_name} with Json success")



from struct import pack, unpack
import binascii
from .serializers.DummyProtocol import DummyProtocol, DummyProtocolData


class Thrift(object):

    BASE_EXCEPTION = {
        'code': 1,
        'message': 2,
        'metadata': 3
    }

    def __init__(self):
        pass

    class TBinaryProtocol(object):

        VERSION_MASK = -65536
        VERSION_1 = -2147418112
        TYPE_MASK = 0x000000ff

        def __init__(self, cl, data: bytes = None, baseException: dict = None):
            self.cl = cl
            self.__last_fid = 0
            self.__last_pos = 0
            self.__last_sid = 0
            self.res = None
            self.data = data
            self.baseException = baseException if baseException is not None else Thrift.BASE_EXCEPTION
            if self.data is not None:
                self.x()

        def readBool(self):
            byte = self.readByte()
            if byte == 0:
                return False
            return True

        def readI16(self):
            buff = self.y(2)
            val, = unpack('!h', buff)
            return val

        def readI32(self):
            buff = self.y(4)
            val, = unpack('!i', buff)
            return val

        def readI64(self):
            buff = self.y(8)
            val, = unpack('!q', buff)
            return val

        def readDouble(self):
            buff = self.y(8)
            val, = unpack('!d', buff)
            return val

        def readBinary(self):
            size = self.readI32()
            s = self.y(size)
            try:
                s = s.decode()
            except:
                pass
            return s

        def readByte(self):
            buff = self.y(1)
            val, = unpack('!b', buff)
            return val

        def readMessageBegin(self):
            sz = self.readI32()
            if sz < 0:
                version = sz & self.VERSION_MASK
                if version != self.VERSION_1:
                    raise Exception(
                        'Bad version in readMessageBegin: %d' % sz)
                _type = sz & self.TYPE_MASK
                name = self.readBinary()
                seqid = self.readI32()
            else:
                raise Exception('Bad version in readMessageBegin: %d' % sz)
            return name, _type, seqid

        def readFieldBegin(self):
            _type = self.readByte()
            if _type == 0:
                return None, _type, 0
            _id = self.readI16()
            return None, _type, _id

        def readMapBegin(self):
            ktype = self.readByte()
            vtype = self.readByte()
            size = self.readI32()
            return ktype, vtype, size

        def readListBegin(self):
            etype = self.readByte()
            size = self.readI32()
            return etype, size

        def writeByte(self, byte):
            buff = pack("!b", byte)
            return buff

        def x(self):
            if len(self.data) < 4:
                return None
            name, _type, seqid = self.readMessageBegin()
            _, ftype, fid = self.readFieldBegin()
            data = None
            d = DummyProtocol()
            if fid == 0:
                data, d = self.z(ftype, fid)
            elif fid == 1:
                error, d = self.z(ftype, fid)
                if ftype == 11:
                    raise Exception(error)
                data = {
                    "error": {
                        "code": error.get(self.baseException['code']),
                        "message": error.get(self.baseException['message']),
                        "metadata": error.get(self.baseException['metadata']),
                        "_data": error
                    }
                }
            else:
                error, d = self.z(ftype, fid)
                raise Exception(f"unknown fid: {fid}, err: {error}, HEX: {self.data.hex()}")
            self.res = data
            self.dummyProtocol = d

        def y(self, num: int):
            data = self.data[self.__last_pos:self.__last_pos + num]
            self.__last_pos += num
            return data

        def z(self, ftype: int, fid: int = None):
            data = None
            dummyProtocol = DummyProtocol(3)
            dummyProtocolData = None
            subType = None
            if ftype == 0:
                pass
            elif ftype == 2:
                data = self.readBool()
            elif ftype == 3:
                data = self.readByte()
            elif ftype == 4:
                data = self.readDouble()
            elif ftype == 6:
                data = self.readI16()
            elif ftype == 8:
                data = self.readI32()
            elif ftype == 10:
                data = self.readI64()
            elif ftype == 11:
                data = self.readBinary()
            elif ftype == 12:
                data = {}
                dummyProtocolData = []
                while True:
                    _, _ftype, _fid = self.readFieldBegin()
                    if _ftype == 0:
                        break
                    data[_fid], _dummyProtocolData = self.z(_ftype, _fid)
                    dummyProtocolData.append(_dummyProtocolData.data)
            elif ftype == 13:
                ktype, vtype, size = self.readMapBegin()
                data = {}
                subType = [ktype, vtype]
                dummyProtocolData = {}
                for i in range(size):
                    _key, _kDPD = self.z(ktype)
                    _val, _vDPD = self.z(vtype)
                    dummyProtocolData[_kDPD.data] = _vDPD.data
                    data[_key] = _val
            elif ftype == 14 or ftype == 15:
                etype, size = self.readListBegin()
                data = []
                dummyProtocolData = []
                subType = [etype]
                for i in range(size):
                    _data, _dummyProtocolData = self.z(etype)
                    data.append(_data)
                    dummyProtocolData.append(_dummyProtocolData.data)
            else:
                raise Exception(f"can't not read type {ftype}")
            if dummyProtocolData is None:
                dummyProtocolData = data
            if dummyProtocolData is not None:
                dummyProtocol.data = DummyProtocolData(
                    fid, ftype, dummyProtocolData, subType)
            return data, dummyProtocol

    class TCompactProtocol(object):

        class TType(object):
            STOP = 0
            VOID = 1
            BOOL = 2
            BYTE = 3
            I08 = 3
            DOUBLE = 4
            I16 = 6
            I32 = 8
            I64 = 10
            STRING = 11
            UTF7 = 11
            STRUCT = 12
            MAP = 13
            SET = 14
            LIST = 15
            UTF8 = 16
            UTF16 = 17

        class CompactType(object):
            STOP = 0x00
            TRUE = 0x01
            FALSE = 0x02
            BYTE = 0x03
            I16 = 0x04
            I32 = 0x05
            I64 = 0x06
            DOUBLE = 0x07
            BINARY = 0x08
            LIST = 0x09
            SET = 0x0A
            MAP = 0x0B
            STRUCT = 0x0C

        CTYPES = {
            TType.STOP: CompactType.STOP,
            TType.BOOL: CompactType.TRUE,  # used for collection
            TType.BYTE: CompactType.BYTE,
            TType.I16: CompactType.I16,
            TType.I32: CompactType.I32,
            TType.I64: CompactType.I64,
            TType.DOUBLE: CompactType.DOUBLE,
            TType.STRING: CompactType.BINARY,
            TType.STRUCT: CompactType.STRUCT,
            TType.LIST: CompactType.LIST,
            TType.SET: CompactType.SET,
            TType.MAP: CompactType.MAP,
        }

        TTYPES = {}
        for k, v in CTYPES.items():
            TTYPES[v] = k
        TTYPES[CompactType.FALSE] = TType.BOOL
        del k
        del v

        def __init__(self, cl, data=None, passProtocol=False, baseException: dict = None):
            self.cl = cl
            self.__last_fid = 0
            self.__last_pos = 0
            self.passProtocol = passProtocol
            self.baseException = baseException if baseException is not None else Thrift.BASE_EXCEPTION
            if data is not None:
                self.data = data
                self.x()

        def getFieldHeader(self, _type, fid):
            delta = fid - self.__last_fid
            res = []
            if 0 < delta <= 15:
                res.append(delta << 4 | _type)
            else:
                res += self.__writeByte(_type)
                res += self.__writeI16(fid)
            self.__last_fid = fid
            return res

        def readVarint(self, data, return_len=False):
            result = 0
            shift = 0
            i = 0
            while True:
                byte = data[i]
                i += 1
                result |= (byte & 0x7f) << shift
                if byte >> 7 == 0:
                    if return_len:
                        return [result, i]
                    return result
                shift += 7

        def writeVarint(self, data):
            out = []
            while True:
                if data & ~0x7f == 0:
                    out.append(data)
                    break
                else:
                    out.append((data & 0xff) | 0x80)
                    data = data >> 7
            return out

        def __writeByte(self, byte):
            return list(pack('!b', byte))

        def __readVarint(self, data, return_len=False):
            return self.readVarint(data, return_len)

        def __readByte(self, data):
            result, = unpack('!b', data[:1])
            return result

        def __readUByte(self, data):
            result, = unpack('!B', data)
            return result

        def __writeI16(self, i16):
            return self.writeVarint(self.makeZigZag(i16, 16))

        def __writeI32(self, i32):
            return self.writeVarint(self.makeZigZag(i32, 32))

        def makeZigZag(self, n, bits):
            # checkIntegerLimits(n, bits)
            return (n << 1) ^ (n >> (bits - 1))

        def fromZigZag(self, n):
            return (n >> 1) ^ -(n & 1)

        def __readZigZag(self, data, return_len=False):
            if return_len:
                (res, len) = self.__readVarint(data, True)
                return [self.fromZigZag(res), len]
            return self.fromZigZag(self.__readVarint(data))

        def readBool(self):
            return self.__bool_value == 0x01

        def __readSize(self, data):
            result = self.__readVarint(data, True)
            if result[0] < 0:
                raise Exception("[__readSize] Length < 0")
            return result

        def readBinary(self, data):
            (size, len) = self.__readSize(data)
            res = data[len:size + len]
            try:
                res = res.decode()
            except:
                pass
            return [res, len + size]

        def __writeUByte(self, byte):
            return list(pack('!B', byte))

        def __writeSize(self, i32):
            return self.writeVarint(i32)

        def readMessageBegin(self):
            proto_id = self.__readUByte(self.y(1))
            if proto_id != 130:
                raise Exception(
                    'Bad protocol id in the message: %d' % proto_id)
            ver_type = self.__readUByte(self.y(1))
            _type = (ver_type >> 5) & 7
            version = ver_type & 15
            if version != 1:
                raise Exception('Bad version: %d (expect %d)' % (version, 1))
            seqid, offset = self.__readVarint(
                self.data[self.__last_pos:], True)
            self.__last_pos += offset
            name, offset = self.readBinary(self.data[self.__last_pos:])
            self.__last_pos += offset
            return name, _type, seqid

        def readFieldBegin(self, data):
            offset = 1
            if len(data) == 0:
                return None, 0, 0, 0
            _type = data[0]
            if _type & 0x0f == 0x00:
                return None, 0, 0, offset
            delta = _type >> 4
            if delta == 0:
                _fid = self.__readI16(data[offset:], True)
                fid = _fid[0]
                offset += _fid[1]
            else:
                fid = self.__last_fid + delta
            self.__last_fid = fid
            _type = _type & 0x0f
            if _type == 0x01:
                self.__bool_value = True
            elif _type == 0x02:
                self.__bool_value = False
            return None, _type, fid, offset

        def readCollectionBegin(self, data):
            size_type = data[0]
            size = size_type >> 4
            _type = size_type & 0x0f
            _len = 0
            if size == 15:
                size, _len = self.__readSize(data[1:])
            return _type, size, _len + 1

        def readMapBegin(self, data):
            size, len = self.__readSize(data)
            types = 0
            if size > 0:
                types = self.__readUByte(data[len:len + 1])
            else:
                return 0, 0, 0, 1  # fixed length
            vtype = types & 0x0f
            ktype = types >> 4
            return ktype, vtype, size, len + 1

        def writeCollectionBegin(self, etype, size):
            a = []
            if size <= 14:
                a += self.__writeUByte(size << 4 | self.CTYPES[etype])
            else:
                a += self.__writeUByte(0xf0 | self.CTYPES[etype])
                a += self.__writeSize(size)
            return a

        def writeMapBegin(self, ktype, vtype, size):
            a = []
            if size == 0:
                a += self.__writeByte(0)
            else:
                a += self.__writeSize(size)
                a += self.__writeUByte(self.CTYPES[ktype]
                                       << 4 | self.CTYPES[vtype])
            return a

        def readDouble(self, data):
            buff = data[:8]
            val, = unpack('<d', buff)
            return val

        def x(self, isFirst=True):
            if isFirst:
                name, _type, seqid = self.readMessageBegin()
            _, ftype, fid, offset = self.readFieldBegin(
                self.data[self.__last_pos:])
            self.__last_pos += offset
            data = None
            d = DummyProtocol()
            if self.passProtocol:
                if ftype == 0:
                    return None
                if isFirst:
                    data = {
                        name: {
                            fid: self.z(ftype, fid)[0]
                        }
                    }
                    while True:
                        data2 = self.x(False)
                        if data2 is None:
                            break
                        data[name].update(data2)
                else:
                    return {
                        fid: self.z(ftype, fid)[0]
                    }
            elif fid == 0:
                data, d = self.z(ftype, fid)
            elif fid == 1:
                error, d = self.z(ftype, fid)
                if ftype == 8:
                    raise Exception(error)
                data = {
                    "error": {
                        "code": error.get(self.baseException['code']),
                        "message": error.get(self.baseException['message']),
                        "metadata": error.get(self.baseException['metadata']),
                        "_data": error
                    }
                }
            else:
                raise Exception(f"unknown fid: {fid}")
            self.res = data
            self.dummyProtocol = d

        def y(self, num: int):
            data = self.data[self.__last_pos:self.__last_pos + num]
            self.__last_pos += num
            return data

        def z(self, ftype: int, fid: int = None):
            data = None
            dummyProtocol = DummyProtocol()
            dummyProtocolData = None
            subType = None
            if ftype == 0:
                pass
            elif ftype == 1:
                data = True
            elif ftype == 2:
                data = False
            elif ftype == 3:
                data = self.readByte(self.data[self.__last_pos:])
                self.__last_pos += 1
            elif ftype == 4:
                data = self.readDouble(self.data[self.__last_pos:])
                self.__last_pos += 8
            elif ftype == 5:
                data, offset = self.readI32(self.data[self.__last_pos:], True)
                self.__last_pos += offset
            elif ftype == 6:
                data, offset = self.readI64(self.data[self.__last_pos:], True)
                self.__last_pos += offset
            elif ftype == 7:
                data = self.readDouble(self.data[self.__last_pos:])
                self.__last_pos += 8
            elif ftype == 8:
                data, offset = self.readBinary(self.data[self.__last_pos:])
                self.__last_pos += offset
            elif ftype == 9 or ftype == 10:
                data = []
                vtype, vsize, vlen = self.readCollectionBegin(
                    self.data[self.__last_pos:])
                self.__last_pos += vlen
                dummyProtocolData = []
                subType = [vtype]
                for _i in range(vsize):
                    _data, _dummyProtocolData = self.z(vtype)
                    data.append(_data)
                    dummyProtocolData.append(_dummyProtocolData.data)
            elif ftype == 11:
                data = {}
                ktype, vtype, size, dlen = self.readMapBegin(
                    self.data[self.__last_pos:])
                self.__last_pos += dlen
                dummyProtocolData = {}
                subType = [ktype, vtype]
                for _i in range(size):
                    _key, _kDPD = self.z(ktype)
                    _val, _vDPD = self.z(vtype)
                    dummyProtocolData[_kDPD.data] = _vDPD.data
                    data[_key] = _val
            elif ftype == 12:
                data = {}
                _dec = Thrift.TCompactProtocol(self.cl)
                dummyProtocolData = []
                while True:
                    _, _ftype, _fid, offset = _dec.readFieldBegin(
                        self.data[self.__last_pos:])
                    self.__last_pos += offset
                    if _ftype == 0:
                        break
                    data[_fid], _dummyProtocolData = self.z(_ftype, _fid)
                    dummyProtocolData.append(_dummyProtocolData.data)
            else:
                raise Exception(f"can't not read type {ftype}")
            if dummyProtocolData is None:
                dummyProtocolData = data
            if subType:
                # FIXED: Unified format
                for i in range(len(subType)):
                    subType[i] = self.TTYPES[subType[i]]
            
            if dummyProtocolData is not None:
                dummyProtocol.data = DummyProtocolData(
                    fid, self.TTYPES[ftype], dummyProtocolData, subType)
            return data, dummyProtocol

        writeByte = __writeByte
        writeI32 = __writeI32
        readByte = __readByte
        __readI16 = __readZigZag
        readI16 = __readZigZag
        readI32 = __readZigZag
        readI64 = __readZigZag
        readSetBegin = readCollectionBegin
        readListBegin = readCollectionBegin

    class TMoreCompactProtocol(TCompactProtocol):
        """
        Author: YinMo (https://github.com/WEDeach)
        Source: CHRLINE (https://github.com/DeachSword/CHRLINE)
        Version: 1.0.12 (令和最新版)
        """

        def __init__(self, cl, a=None, baseException: dict = None, readWith: str = None):
            self.cl = cl                # cl  init
            self.__a = []               # 1st init
            self.__b = []               # 1st init
            self.__c = self._b          # 1st init
            self.__d = []               # 2nd init
            self.__e = []               # 2nd init
            self.__f = []               # 3rd init
            self.__h = self._c          # 2nd init
            self.__last_fid = 0         # base fid
            self.__last_pos = 0         # base pos
            self.__last_sid = 0         # base sid
            self._a()                   # 4th init
            self.res = None             # base res
            self.baseException = baseException              # init
            if baseException is None:                       #
                self.baseException = Thrift.BASE_EXCEPTION  # base
            self.readWith = readWith    # readWith
            if a is not None:           # not None
                self.d(a)               # for data

        def a(self, cArr, b2):
            self.__b[b2] = cArr         # bk array!!
            i2 = 0                      # base init!
            for c2 in cArr:             #
                if c2 == '0':           # is 0
                    i2 = (i2 << 1) + 1  # + 1
                elif c2 == '1':         # is 1
                    i2 = (i2 << 1) + 2  # + 2
            self.__a[i2] = b2           # init array

        def b(self):
            i2 = 0                                  # base init
            i3 = 0                                  # base init
            while True:                             #
                l2 = self.data[self.__last_pos]     # so good!!
                self.__last_pos += 1                # fixed pos
                i2 |= (l2 & 127) << i3              # yea baby!
                if (l2 & 128) != 128:               # come on!!
                    return i2                       # break!!!!
                i3 += 7                             # + 7!!!!!!

        def c(self, p, i2):
            if (i2 == 0):               # is 0!!
                return []               # break!
            bArr = self.data[p:p + i2]  # read!!
            return list(bArr)           # break!

        def d(self, d):
            self.data = d       # base init!
            return self.t()     # base init?

        def e(self):
            a = None                # base init
            b = None                # base init
            c = 0                   # base init
            d = DummyProtocol()     # dummy >w<
            _fid = self.b()         # read fid!
            if _fid == 0:           #
                pass                #
            elif _fid in [1, 2]:                #
                fid, = self.n(_fid)             # read
                if fid == 0:                    #
                    _type = self.w()            # read data
                    a, d = self.g(_type, fid)   # read data
                elif fid == 1:                  #
                    _type = self.w()            # read data
                    a, d = self.g(_type, fid)   # read data
                    a = {                                                       #
                        "error": {                                              #
                            "code": a.get(self.baseException['code']),          #
                            "message": a.get(self.baseException['message']),    #
                            "metadata": a.get(self.baseException['metadata']),  #
                            "_data": a                                          #
                        }                                                       #
                    }                                                           #
                elif fid == 5:                                      #
                    _type = self.w()                                # read data
                    a, d = self.g(_type, fid)                       # exception!
                    raise Exception(a)                              # raise!
                else:                                               #
                    raise EOFError(f"fid {fid} not implemented")    # exception!
            else:                                                           #
                _type = self.w()                                            # read data
                a, d = self.g(_type)                                        # read data
                raise EOFError(                                             #
                    f"recv fid `{_fid}`, expected `1`, message: `{a}`")     # err
            self.res = a            # write data
            self.dummyProtocol = d  # write data

        def f(self, n):
            return (n >> 1) ^ -(n & 1)  # hmm...

        def g(self, t, fid=None):
            a = None                            # base
            b = None                            # base
            c = 0                               # base
            dummyProtocol = DummyProtocol()     # base
            dummyProtocolData = None            # base
            subType = None                      # base
            if t == 2:          #
                b = self.b()    # read
                a = bool(b)     # bool
            elif t == 3:                                            #
                dec = Thrift.TCompactProtocol(self.cl)              # init
                a = dec.readByte(self.data[self.__last_pos:])       # byte
                self.__last_pos += 1                                # fix!
            elif t == 4:                                            #
                dec = Thrift.TCompactProtocol(self.cl)              # init
                a = dec.readDouble(self.data[self.__last_pos:])     # read
                self.__last_pos += 8                                # fix!
            elif t == 8:                                    #
                _a = self.x(self.data[self.__last_pos:])    # read
                a = self.f(_a)                              # int!
            elif t == 10:       #
                _a = self.b()   # read
                a = self.f(_a)  # int?
            elif t == 11:       #
                a = self.s()    # str!
            elif t == 12:               #
                a = {}                  # base
                b = self.b()            # read
                c = self.n(b)           # read
                dummyProtocolData = []  # base
                for d in c:                                             #
                    a[d], _dummyProtocolData = self.g(                  #
                        self.w(), d)                                    # fld!
                    dummyProtocolData.append(_dummyProtocolData.data)   # init
            elif t == 13:               #
                a = {}                  # base
                c = self.b()            # read
                subType = [0, 0]        # base
                dummyProtocolData = {}  # base
                if c != 0:              #
                    d = self.y()        # read
                    t1, t2 = self.q(d)  # read
                    subType = [t1, t2]  # init
                    for i in range(c):                              #
                        k, _kDPD = self.g(t1)                       # key!
                        v, _vDPD = self.g(t2)                       # val!
                        dummyProtocolData[_kDPD.data] = _vDPD.data  #
                        a[k] = v                                    # dict
            elif t == 14 or t == 15:                                #
                a = []                                              # base
                dec = Thrift.TCompactProtocol(self.cl)              # init
                ftype, count, offset = dec.readCollectionBegin(     #
                    self.data[self.__last_pos:])                    # read
                self.__last_pos += offset                           # fix!
                subType = [self._d(ftype)]                          # init
                dummyProtocolData = []                              # base
                for i in range(count):                                  #
                    b, _dummyProtocolData = self.g(                     #
                        self._d(ftype))                                 # read
                    a.append(b)                                         # list
                    dummyProtocolData.append(_dummyProtocolData.data)   # init
            elif t == 16:                       #
                b = self.b()                    # read
                c = -(b & 1) ^ self._e(b, 1)    # wtf?
                d = c + self.__last_sid         # fix?
                self.__last_sid = d             # idk.
                a = str(d)                      # str!
                t = 11                          # fix.
            elif t == 17:               #
                b = self.b()            # read
                if len(self.__e) > b:               #
                    a = self.__e[b]                 # str?
                    t = 11                          # fix.
                else:                               #
                    print(f"mid not found: {b}")    # no way
            else:                                           #
                raise Exception(f"cAN't rEad TyPE: {t}")    # err!
            if dummyProtocolData is None:                   #
                dummyProtocolData = a                       # base
            dummyProtocol.data = DummyProtocolData(         #
                fid, t, dummyProtocolData, subType)         # good
            return a, dummyProtocol                         # nice

        def h(self, n):
            return (n << 1) ^ (n >> 31)     # hmm...

        def m(self):
            a = self.b()                                                    # get count
            for _a in range(a):                                             #
                bArr = [self.data[self.__last_pos]]                         # coooooool
                bArr += self.__h(                                           #
                    self.data[self.__last_pos + 1:self.__last_pos + 17])    # not magic
                self.__e.append(bytes(bArr).decode())                       # wow magic
                self.__last_pos += 17                                       # real pos?
            self.e()                                                        # base init

        def n(self, d):
            a = []                  # base init
            i = 0                   # base init
            while True:             #
                b = 1 << i          # set &
                if b > d:           #
                    break           # break
                elif d & b != 0:    #
                    a.append(i)     # add
                i += 1              # + 1
            return a                # break

        def q(self, d):
            return self._d(d >> 4), self._d(d & 15)  # cool

        def s(self):
            a = self.b()                                        # read value
            b = self.data[self.__last_pos:self.__last_pos + a]  # init first
            try:                    #
                b = b.decode()      # any ideas?
            except:                 #
                pass                # lamo idea.
            self.__last_pos += a    # fixed pos!
            return b                # - break! -

        def t(self):
            self.__last_pos = 3                                 # fixed pos
            if len(self.data) == 4:                             #
                raise Exception(                                #
                    f"Invalid data: {self.data} (code: 20)")    # raise
            a = self.b()                    # first data
            b = self.c(self.__last_pos, a)  # 2nd data!!
            self.__d = list(bytes(a << 1))  # 3rd? no!!!
            d = 0                           # base init
            e = 0                           # base init
            f = 0                           # base init
            g = 0                           # base init
            for h in b:                                         #
                _a = 0                                          # base value!
                _b = 128                                        # base value?
                while _a < 8:                                   #
                    if h & _b == 0:                             #
                        d = (g << 1) + 1                        # + 1
                    else:                                       #
                        d = (g << 1) + 2                        # + 2
                    if self.__a[d] != 0:                        #
                        if f >= len(self.__d):                  #
                            self.__d += [len(self.__d)] * 4     # x 4
                        self.__d[f] = self.__a[d]               # set
                        f += 1                                  # + 1
                        g = 0                                   # = 0
                    else:                                       #
                        g = d                                   # set!
                    _b >>= 1                                    # move
                    _a += 1                                     # + 1!
            self.__last_pos += a                                # fixed pos
            self.m()                                            # base init

        def w(self):
            a = self.__d[self.__last_fid]   # read!
            self.__last_fid += 1            # + 1!!
            return a                        # break

        def x(self, a, b=False):
            c = 0                           # base init
            d = 0                           # base init
            i = 0                           # base init
            while True:                     #
                e = a[i]                    # read
                i += 1                      # + 1!
                c |= (e & 0x7f) << d        # move
                if e >> 7 == 0:             #
                    self.__last_pos += i    # + i!
                    if b:                   #
                        return [c, i]       # break
                    return c                # break
                d += 7                      # + 7!!

        def y(self):
            a = self.data[self.__last_pos]  # read!
            self.__last_pos += 1            # + 1!!
            return a                        # break

        def z(self):
            if len(self.data) > self.__last_pos:    # Next?
                return True                         # True!
            return False                            # False

        def _a(self):
            self.__a = list(bytes(512))                             # base init
            self.__b = list(bytes(18))                              # base init
            self.__c(['1', '0', '1', '1'], 2)                       # cool yea?
            self.__c(['1', '0', '1', '0', '1', '0', '0', '1'], 3)   # idk why..
            self.__c(['1', '0', '1', '0', '1', '0', '0', '0'], 4)   # too long!
            self.__c(['1', '0', '1', '0', '1', '1', '1'], 6)        # plz make!
            self.__c(['0', '1'], 8)                                 # ez plz!!!
            self.__c(['0', '0'], 10)                                # no! 0 & 0
            self.__c(['1', '0', '1', '0', '0'], 11)                 # what? bin
            self.__c(['1', '1', '0', '1'], 12)                      # stop it!!
            self.__c(['1', '0', '1', '0', '1', '1', '0'], 13)       # aaaaaaaaa
            self.__c(['1', '0', '1', '0', '1', '0', '1'], 14)       # AaAAaaaAa
            self.__c(['1', '1', '0', '0'], 15)                      # * DIED! *
            self.__c(['1', '1', '1'], 16)                           # 1 & 1 & 1
            self.__c(['1', '0', '0'], 17)                           # 1 & 0 & 0

        def _b(self, cArr, b2):
            self.__b[b2] = cArr         # base init
            i2 = 0                      # base init
            for c2 in cArr:             #
                if c2 == '0':           #
                    i2 = (i2 << 1) + 1  # + 1!!
                elif c2 == '1':         #
                    i2 = (i2 << 1) + 2  # + 2!!
            self.__a[i2] = b2           # break

        def _c(self, val):
            return binascii.b2a_hex(val)  # magic right?

        def _d(self, val):
            if val == 0:        #
                return 0        # break
            if val in [1, 2]:   #
                return 2        # break
            if val == 3:        #
                return 3        # break
            if val == 4:        #
                return 6        # break
            if val == 5:        #
                return 8        # break
            if val == 6:        #
                return 10       # break
            if val == 7:        #
                return 4        # break
            if val == 8:        #
                return 11       # break
            if val == 9:        #
                return 15       # break
            if val == 10:       #
                return 14       # break
            if val == 11:       #
                return 13       # break
            if val == 12:       #
                return 12       # break
            raise Exception(f'Invalid type: {val}') # error

        def _e(self, val, n):
            if val >= 0:                                    #
                val >>= n                                   # >>=?
            else:                                           #
                val = ((val + 0x10000000000000000) >> n)    # wtf?
            return val                                      # ret?


def checkIntegerLimits(i, bits):
    if bits == 8 and (i < -128 or i > 127):
        raise Exception('INVALID_DATA',
                        "i8 requires -128 <= number <= 127")
    elif bits == 16 and (i < -32768 or i > 32767):
        raise Exception('INVALID_DATA',
                        "i16 requires -32768 <= number <= 32767")
    elif bits == 32 and (i < -2147483648 or i > 2147483647):
        raise Exception('INVALID_DATA',
                        "i32 requires -2147483648 <= number <= 2147483647")
    elif bits == 64 and (i < -9223372036854775808 or i > 9223372036854775807):
        raise Exception('INVALID_DATA',
                        "i64 requires -9223372036854775808 <= number <= 9223372036854775807")